import datetime
import os
from typing import TYPE_CHECKING, Callable, Final, NamedTuple, Optional

import rich

from celline.DB.dev.handler import HandleResolver
from celline.DB.dev.model import SampleSchema
from celline.config import Config, Setting
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleInfo, SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager

if TYPE_CHECKING:
    from celline import Project


class Integrate(CellineFunction):
    class JobContainer(NamedTuple):
        nthread: str
        cluster_server: str
        jobname: str
        logpath: str
        r_path: str
        exec_root: str
        sample_ids: str
        project_ids: str
        all_bcmat_path: str
        all_data_sample_dir_path: str
        outfile_path: str
        logpath_runtime: str
        project_name: str

    def __init__(
        self,
        filter_func: Optional[Callable[[SampleSchema], bool]],
        outfile_name: Optional[str],
    ) -> None:
        self.filter_func = filter_func
        self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name
        if outfile_name is None:
            self.outfile_path: str = (
                f"{Config.PROJ_ROOT}/integration/integrated_{datetime.datetime.now()}"
            )
        else:
            self.outfile_path = f"{Config.PROJ_ROOT}/integration/{outfile_name}"

    def register(self) -> str:
        return "integrate"

    def call(self, project: "Project"):
        now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        target_samples: list[SampleInfo] = []
        for info in SampleResolver.samples.values():
            if info.path.is_counted:
                if self.filter_func is None:
                    add = True
                else:
                    add = self.filter_func(info.schema)
                if add:
                    target_samples.append(info)
            else:
                rich.print(
                    f":warning: [bold yellow]Warning[/] Target sample {info.schema.key} is not counted or preprocessed yet."
                )
        os.makedirs(f"{Config.PROJ_ROOT}/integration", exist_ok=True)
        os.makedirs(f"{Config.PROJ_ROOT}/integration/logs", exist_ok=True)
        TemplateManager.replace_from_file(
            file_name="integrate.sh",
            structure=Integrate.JobContainer(
                nthread=str(1),
                cluster_server=""
                if self.cluster_server is None
                else self.cluster_server,
                jobname="Integration",
                logpath=f"{Config.PROJ_ROOT}/integration/logs/integrate_{now}.log",
                r_path=f"{Setting.r_path}script",
                exec_root=Config.EXEC_ROOT,
                sample_ids=",".join(
                    [str(sample.schema.key) for sample in target_samples]
                ),
                project_ids=",".join(
                    [
                        sample.schema.parent if sample.schema.parent is not None else ""
                        for sample in target_samples
                    ]
                ),
                all_bcmat_path=",".join(
                    [
                        f"{sample.path.resources_sample_counted}/outs/filtered_feature_bc_matrix.h5"
                        for sample in target_samples
                    ]
                ),
                all_data_sample_dir_path=",".join(
                    [f"{sample.path.data_sample}" for sample in target_samples]
                ),
                outfile_path=self.outfile_path,
                logpath_runtime=f"{Config.PROJ_ROOT}/integration/logs/RUNTIME_integrate_{now}.log",
                project_name=f"{Setting.name}",
            ),
            replaced_path=f"{Config.PROJ_ROOT}/integration/integrate_{now}.sh",
        )
        ThreadObservable.call_shell(
            [f"{Config.PROJ_ROOT}/integration/integrate_{now}.sh"]
        ).watch()
        return project



from dataclasses import dataclass
import datetime
import os
import sys
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Final,
    List,
    NamedTuple,
    Optional,
    Union,
)

import polars as pl
import rich
from rich.progress import track
import toml
import tqdm

from celline.DB.dev.handler import HandleResolver
from celline.DB.dev.model import SampleSchema
from celline.config import Config, Setting
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleInfo, SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager
from celline.utils.serialization import NamedTupleAndPolarsStructure

if TYPE_CHECKING:
    from celline import Project


class BatchCorrection(CellineFunction):
    """
    Correction batch effect
    """

    def __init__(
        self,
        output_file_path: str,
        filter_func: Optional[Callable[[SampleSchema], bool]],
    ) -> None:
        """
        #### Add accession ID to DB & your project.

        #### Note: Parallel calculations are not supported

        Args:
            h5ad_path (<List[Add.SampleInfo]> | <pl.DataFrame>): Accession ID to add.
        """
        self.filter_func = filter_func
        self.output_file_path = output_file_path
        self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name

    def register(self) -> str:
        return "batch"

    def call(self, project: "Project") -> "Project":
        """
        Call the function to add accession IDs to the project.

        Args:
            project (<Project>): The project to add the accession IDs to.

        Returns:
            <Project>: The project with the added accession IDs.
        """
        # STEP0: Data collection
        target_samples: list[SampleInfo] = []
        for info in SampleResolver.samples.values():
            if info.path.is_counted:
                if self.filter_func is None:
                    add = True
                else:
                    add = self.filter_func(info.schema)
                if add:
                    target_samples.append(info)
            else:
                rich.print(
                    f":warning: [bold yellow]Warning[/] Target sample {info.schema.key} is not counted or preprocessed yet."
                )
        os.makedirs(f"{Config.PROJ_ROOT}/batch", exist_ok=True)
        os.makedirs(f"{Config.PROJ_ROOT}/batch/logs", exist_ok=True)
        os.makedirs(f"{Config.PROJ_ROOT}/batch/logs/runtime", exist_ok=True)
        # [OBS: STEP1]: Pure integrate
        # self.__pure_merge(target_samples)
        # [OBS: STEP2]: Convert rds to h5ad
        # self.__convert(project)
        # STEP3: Correct batch effect
        self.__correct_batch_effect(target_samples)

        return project

    def __correct_batch_effect(self, target_samples: list[SampleInfo]):
        class JobContainer(NamedTuple):
            nthread: str
            cluster_server: str
            logpath: str
            r_path: str
            exec_root: str
            output_dir: str
            sample_ids: str
            project_ids: str
            logpath_runtime: str
            proj_path: str

        NOW: Final[str] = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        TemplateManager.replace_from_file(
            file_name="batch_cor.sh",
            structure=JobContainer(
                nthread=str(1),
                cluster_server=""
                if self.cluster_server is None
                else self.cluster_server,
                logpath=f"{Config.PROJ_ROOT}/batch/logs/removebatch_{NOW}.log",
                r_path=f"{Setting.r_path}script",
                exec_root=Config.EXEC_ROOT,
                proj_path=Config.PROJ_ROOT,
                sample_ids=",".join(
                    [str(sample.schema.key) for sample in target_samples]
                ),
                project_ids=",".join(
                    [
                        sample.schema.parent if sample.schema.parent is not None else ""
                        for sample in target_samples
                    ]
                ),
                output_dir=self.output_file_path,
                logpath_runtime=f"{Config.PROJ_ROOT}/batch/logs/removebatch_{NOW}.sh",
            ),
            replaced_path=f"{Config.PROJ_ROOT}/batch/removebatch_{NOW}.sh",
        )
        ThreadObservable.call_shell(
            [f"{Config.PROJ_ROOT}/batch/removebatch_{NOW}.sh"]
        ).watch()



from dataclasses import dataclass
import os
import shutil
import sys
from typing import TYPE_CHECKING, Final, List, NamedTuple, Optional

import pandas as pd
import polars as pl
import rich
from rich.console import Console
from rich.table import Table

from celline.DB.dev.handler import HandleResolver
from celline.DB.dev.model import SampleSchema
from celline.config import Config, Setting
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager
from celline.utils.path import Path

if TYPE_CHECKING:
    from celline import Project


@dataclass
class CellTypeModel:
    species: str
    suffix: Optional[str]


class BuildCellTypeModel(CellineFunction):
    """### Build cell type prediction model"""

    class JobContainer(NamedTuple):
        """
        Represents job information for data download.
        """

        nthread: str
        cluster_server: str
        jobname: str
        logpath: str
        h5matrix_path: str
        celltype_path: str
        dist_dir: str
        r_path: str
        exec_root: str

    def __init__(
        self,
        species: str,
        suffix: str,
        nthread: int,
        h5matrix_path: str,
        celltype_path: str,
    ) -> None:
        if not celltype_path.endswith(".tsv"):
            rich.print(
                "[bold red]Build Error[/] celltype_path should be .tsv file path."
            )
            self.__show_help()
            sys.exit(1)
        _df = pl.read_csv(celltype_path, separator="\t")
        if _df.columns != ["cell", "celltype"]:
            rich.print(
                "[bold red]Build Error[/] celltype dataframe should be composed of cell, celltype column."
            )
            self.__show_help()
            sys.exit(1)
        if (
            not h5matrix_path.endswith(".h5")
            and not h5matrix_path.endswith(".loom")
            and not h5matrix_path.endswith(".h5seurat")
            and not h5matrix_path.endswith(".h5seuratv5")
        ):
            rich.print(
                "[bold red]Build Error[/] h5matrix_path should be .h5, h5seurat, h5seuratv5 or .loom file path."
            )
        self.model: Final[CellTypeModel] = CellTypeModel(species, suffix)
        self.nthread: Final[int] = nthread
        self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name
        self.h5matrix_path: Final[str] = h5matrix_path
        self.celltype_path: Final[str] = celltype_path

    def __show_help(self):
        df = pd.DataFrame(
            {
                "cell": [
                    "10X82_2_TCTCTCACCAGTTA",
                    "10X82_2_TCTCTCACCAGTTC",
                    "10X82_2_TCTCTCACCAGTTT",
                ],
                "celltype": ["Astrocyte", "Oligodendrocyte", "Neuron"],
            },
            index=None,
        )
        table = Table(show_header=True, header_style="bold magenta")
        console = Console()
        for column in df.columns:
            table.add_column(column)
        for _, row in df.iterrows():
            table.add_row(*row.astype(str).tolist())
        rich.print(
            """
[bold green]:robot: How to use?[/]

* [bold]h5matrix_path<str>[/]: h5 matrix path. This data should be h5 matrix which be output from Cellranger.
* [bold]celltype_path<str>[/]: cell type path. This dataframe should be tsv format which have following dataframe structure."""
        )
        console.print(table)

    def call(self, project: "Project"):
        dist_dir = f"{Config.PROJ_ROOT}/reference/{self.model.species.replace(' ', '_')}/{self.model.suffix if self.model.suffix is not None else 'default'}"
        if (
            os.path.isdir(dist_dir)
            and not os.path.isfile(f"{dist_dir}/reference.pred")
            and not os.path.isfile(f"{dist_dir}/reference.h5seurat")
        ):
            shutil.rmtree(dist_dir)
        os.makedirs(dist_dir, exist_ok=True)
        TemplateManager.replace_from_file(
            file_name="build_reference.sh",
            structure=BuildCellTypeModel.JobContainer(
                nthread=str(self.nthread),
                cluster_server=""
                if self.cluster_server is None
                else self.cluster_server,
                jobname="BuildCelltypeModel",
                logpath=f"{dist_dir}/build.log",
                h5matrix_path=self.h5matrix_path,
                dist_dir=dist_dir,
                celltype_path=self.celltype_path,
                r_path=f"{Setting.r_path}script",
                exec_root=Config.EXEC_ROOT,
            ),
            replaced_path=f"{dist_dir}/build.sh",
        )
        ThreadObservable.call_shell([f"{dist_dir}/build.sh"]).watch()
        return project


class PredictCelltype(CellineFunction):
    class JobContainer(NamedTuple):
        """
        Represents job information for data download.
        """

        nthread: str
        cluster_server: str
        jobname: str
        logpath: str
        r_path: str
        exec_root: str
        reference_seurat: str
        reference_celltype: str
        projects: str
        samples: str
        resources_path: str
        data_path: str

    def __init__(self, model: CellTypeModel, re_predict: bool = False) -> None:
        self.model = model
        self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name
        self.re_predict: Final[bool] = re_predict

    def register(self) -> str:
        return "predict_celltype"

    def call(self, project: "Project"):
        __dist_dir = f"{Config.PROJ_ROOT}/reference/{self.model.species.replace(' ', '_')}/{self.model.suffix if self.model.suffix is not None else 'default'}"
        refh5 = f"{__dist_dir}/reference.h5seurat"
        refpred = f"{__dist_dir}/reference.pred"
        sample_infos = [
            sample
            for sample in SampleResolver.samples.values()
            if sample.schema.species.replace(" ", "_")
            == self.model.species.replace(" ", "_")
        ]
        if not self.re_predict:
            sample_infos = [
                sample
                for sample in sample_infos
                if not sample.path.is_predicted_celltype
            ]

        TemplateManager.replace_from_file(
            file_name="predict_celltype.sh",
            structure=PredictCelltype.JobContainer(
                nthread=str(1),
                cluster_server=""
                if self.cluster_server is None
                else self.cluster_server,
                jobname="PredictCelltype",
                logpath=f"{Config.PROJ_ROOT}/reference/prediction.log",
                reference_seurat=refh5,
                reference_celltype=refpred,
                r_path=f"{Setting.r_path}script",
                exec_root=Config.EXEC_ROOT,
                projects=",".join(
                    [
                        sample.schema.parent
                        for sample in sample_infos
                        if sample.schema.parent is not None
                    ]
                ),
                samples=",".join(
                    [
                        str(sample.schema.key)
                        for sample in sample_infos
                        if sample.schema.parent is not None
                    ]
                ),
                resources_path=f"{Config.PROJ_ROOT}/resources",
                data_path=f"{Config.PROJ_ROOT}/data",
            ),
            replaced_path=f"{Config.PROJ_ROOT}/reference/prediction.sh",
        )
        ThreadObservable.call_shell(
            [f"{Config.PROJ_ROOT}/reference/prediction.sh"]
        ).watch()
        return project



import argparse
import os
from typing import TYPE_CHECKING, Dict, Final, Optional

import polars as pl
import scanpy as sc
import scrublet as scr
import toml
from rich.console import Console

from celline.config import Config
from celline.DB.dev.handler import HandleResolver
from celline.DB.dev.model import SampleSchema
from celline.functions._base import CellineFunction
from celline.utils.path import Path

if TYPE_CHECKING:
    pass

console = Console()

class Preprocess(CellineFunction):
    TARGET_CELLTYPE: Final[Optional[list[str]]]
    def __init__(self, target_celltype: Optional[list[str]] = None):
        """
        Initialize the Preprocess class.

        This constructor sets up the Preprocess object with an optional list of target cell types.

        Parameters:
        -----------
        target_celltype : Optional[list[str]], default=None
            A list of target cell types to be considered in the preprocessing.
            If None, all cell types will be considered.

        Returns:
        --------
        None
        """
        self.TARGET_CELLTYPE = target_celltype

    def call(self, project):
        """
        Perform preprocessing on the given project's samples.

        This function reads sample information from a TOML file, processes each sample,
        performs quality control, and generates cell information for further analysis.

        Parameters:
        -----------
        project : object
            The project object containing information about the samples to be processed.

        Returns:
        --------
        project : object
            The input project object, potentially modified during processing.

        Raises:
        -------
        ReferenceError
            If a sample ID cannot be resolved.
        KeyError
            If a sample's parent information is missing.
        """
        sample_info_file = f"{Config.PROJ_ROOT}/samples.toml"
        if not os.path.isfile(sample_info_file):
            print("sample.toml could not be found. Skipping.")
            return project
        with open(sample_info_file, mode="r", encoding="utf-8") as f:
            samples: Dict[str, str] = toml.load(f)
            for sample_id in samples:
                resolver = HandleResolver.resolve(sample_id)
                if resolver is None:
                    raise ReferenceError(
                        f"Could not resolve target sample id: {sample_id}"
                    )
                sample_schema: SampleSchema = resolver.sample.search(sample_id)
                if sample_schema.parent is None:
                    raise KeyError("Could not find parent")
                path = Path(sample_schema.parent, sample_id)
                path.prepare()
                if path.is_counted and path.is_predicted_celltype:
                    adata = sc.read_10x_h5(path.resources_sample_counted)
                    obs = (
                        pl.DataFrame(adata.obs.reset_index())
                        .rename({"index": "barcode"})
                        .with_columns(pl.lit(sample_schema.parent).alias("project"))
                        .with_columns(pl.lit(sample_id).alias("sample"))
                        .with_columns((pl.concat_str(pl.col("sample"), pl.cum_count("sample"), separator="_")).alias("cell"))
                        .join(
                            pl.read_csv(
                            path.data_sample_predicted_celltype,
                            separator="\t",
                            ).rename({"scpred_prediction": "cell_type"}),
                            on="cell",
                        )
                        .with_columns(
                            (pl.col("cell_type").is_in(obs.select(pl.col("cell_type")).unique().get_column("cell_type").to_list() if self.TARGET_CELLTYPE is None else self.TARGET_CELLTYPE)).alias("include")
                        )
                    )
                    scrub = scr.Scrublet(adata.X)
                    doublet_scores, predicted_doublets = scrub.scrub_doublets(verbose=False)
                    adata.obs["doublet_score"] = doublet_scores
                    adata.obs["predicted_doublets"] = predicted_doublets
                    mt_prefix = "mt-"
                    adata.var["mt"] = adata.var_names.str.startswith(mt_prefix)
                    sc.pp.calculate_qc_metrics(
                        adata,
                        qc_vars=["mt"],
                        percent_top=None,
                        log1p=False,
                        inplace=True,
                    )
                    (
                        obs
                        .with_columns(((pl.col("n_genes_by_counts") >= 200) & pl.col("include")).alias("include"))
                        .with_columns(((pl.col("n_genes_by_counts") <= 5000) & pl.col("include")).alias("include"))
                        .with_columns(((pl.col("pct_counts_mt") <= 5) & pl.col("include")).alias("include"))
                        .with_columns(((pl.col("predicted_doublets") == False) & pl.col("include")).alias("include"))
                        .write_csv(f"{path.data_sample}/cell_info.tsv", separator="\t")
                    )

    def add_cli_args(self, parser: argparse.ArgumentParser) -> None:
        """Add CLI arguments for the Preprocess function."""
        parser.add_argument(
            '--target-celltype', '-t',
            nargs='+',
            help='Target cell types to include in preprocessing'
        )

    def cli(self, project, args: Optional[argparse.Namespace] = None):
        """CLI entry point for Preprocess function."""
        target_celltype = None
        if args and hasattr(args, 'target_celltype'):
            target_celltype = args.target_celltype
            
        console.print("[cyan]Starting preprocessing...[/cyan]")
        if target_celltype:
            console.print(f"Target cell types: {', '.join(target_celltype)}")
        
        preprocess_instance = Preprocess(target_celltype)
        return preprocess_instance.call(project)

    def get_description(self) -> str:
        return """Preprocess counted data with quality control and cell type filtering.
        
This function performs quality control on counted data, calculates QC metrics,
detects doublets, and filters cells based on specified criteria."""

    def get_usage_examples(self) -> list[str]:
        return [
            "celline run preprocess",
            "celline run preprocess --target-celltype Neuron Astrocyte"
        ]


import os
from typing import NamedTuple, Type, Optional
from celline.DB.dev.model import BaseModel, Primary, BaseSchema
from pprint import pprint

from dataclasses import dataclass

@dataclass
class Transcriptome_Schema(BaseSchema):
    built_path: str

class Transcriptome(BaseModel[Transcriptome_Schema]):

    def set_class_name(self) -> str:
        return __class__.__name__

    def def_schema(self) -> Type[Transcriptome_Schema]:
        return Transcriptome_Schema

    def search(self, acceptable_id: str, force_search=False) -> Optional[str]:
        target = self.get(Transcriptome_Schema, lambda d: d.key == acceptable_id)
        if len(target) > 0:
            return target[0].built_path
        return None


    def add_path(self, species: str, built_path: str, force_update=True):
        obj = Transcriptome()
        if not os.path.isdir(built_path):
            print(f"Built_path does not exist. {built_path}")
        if (self.search(species) is not None) and (not force_update):
            print(f"Transcriptome of {species} is already exists.")
            return
        obj.add_schema(Transcriptome_Schema(key=species, parent=None, children=None, title=species, built_path=built_path))
